import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='hueyx',
    version='1.0.3-dev2',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',  # example license
    description='Django huey extension which supports multiple huey queues.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/APGSGA/hueyx',
    author='Severin Bühler',
    author_email='severin.buehler@apgsga.ch',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'cached-property',
        'huey>=2.3.0',
        'redis',
        'python-redis-lock',
    ],
)
